// Generation.h: interface for the Generation class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_GENERATION_H__8FA8C9D4_8B21_11D3_9887_0040056D374E__INCLUDED_)
#define AFX_GENERATION_H__8FA8C9D4_8B21_11D3_9887_0040056D374E__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include	<afxtempl.h>
#include	"Random.h"

// Warning!  These numbers are NOT chosen arbitrarily. They must be
// related by POP_SIZE choose 2 = PARENT_SIZE.  That is, the formula:
//	               PARENT_SIZE!
//	POP_SIZE = -----------------------
//	           2! * (PARENT_SIZE - 2)!
#define	POP_SIZE	105
#define	PARENT_SIZE	15

// The Generation object holds a big array of T objects and provides
// methods for creating and advancing a population

template <class T>
class Generation
{
public:
	Generation();
	~Generation();

	void ChooseParents(void);
	int operator++(int);
	void SetMutationRate(int rate);
	T GetBest(void);
	void Init(const CArray<CPoint, CPoint&>& ptarray);

	T		bestEver;					// The best member EVER seen

private:
	int		genCnt;

	int		Fitness[POP_SIZE];			// All member fitnesses
	int		mutationRate;				// Labels changed per generation
	int		pointCount;					// Labels per map

	int		Parent[PARENT_SIZE];		// Array of indexes into the population
	T		evenPopulation[POP_SIZE];	// The members of the population
	T		oddPopulation[POP_SIZE];

	int		bestEverFitness;			// The fitness of best EVER

	int		bestOfBreed;				// The fittest in THIS generation

	Random<int>		randomMember;
	Random<int>		randomPoint;
	Random<int>		randomPlacement;
};

template <class T>
Generation<T>::Generation()
		: genCnt(0), bestOfBreed(0), bestEverFitness(0),
		  randomMember(0, POP_SIZE-1),
		  randomPlacement(0, 3)			{}

template <class T>
Generation<T>::~Generation()			{}


// There are going to be millions of generations--so performance counts.
// I do every reasonable thing to make each generation faster.
// Here's the general scheme:
// Look through the generation for the fittest members.  Don't copy
// them into an array of parent members (too slow), instead, just
// remember where the good ones are (save the index into parent array).
// Then use these parents to create new members into the other (odd
// or even) generation.  This double generation scheme means the parents
// don't have to be saved from the destruction of overwriting.  Since
// the members have to be written anyway, there is only a space penalty
// but no performance penalty.

// Throughout the generation code there is heavy repetition.
// Almost everything is done for both the even and odd generations.
// The conditionals (even or odd) could have been checked inside the
// loops to make the code much smaller, but performance matters here.

template <class T>
void Generation<T>::ChooseParents(void)
{
	if ((genCnt % 2) == 0)
	{
		for (int i=0; i<POP_SIZE; i++)
			Fitness[i] = evenPopulation[i].DetermineFitness();
	}
	else
	{
		for (int i=0; i<POP_SIZE; i++)
			Fitness[i] = oddPopulation[i].DetermineFitness();
	}

	// To find the fittest member in the population, look at
	// the fitness of the first grid, make that a reference,
	// then compare every other grid to it.  When another is
	// fitter, use that as the reference.
	bestOfBreed = Fitness[0];
	int index = 0;
	for (int i=1; i<POP_SIZE; i++)
	{
		if (Fitness[i] > bestOfBreed)
		{
			bestOfBreed = Fitness[i];
			index = i;
		}
	}
	// If this is a new record, remember it:
	if (bestOfBreed >= bestEverFitness)
	{
		bestEverFitness = bestOfBreed;
		if ((genCnt % 2) == 0)
			bestEver = evenPopulation[index];
		else
			bestEver = oddPopulation[index];
	}
	Parent[0] = index;					// Remember where this parent was
	Fitness[index] = -1;				// Don't pick it again

	// Now find the next fittest members to use as parents:
	int nextBiggest;
	for (int p=1; p<PARENT_SIZE; p++)
	{
		// Find the biggest in the fitness array:
		nextBiggest = Fitness[0];
		index = 0;
		for (int i=1; i<POP_SIZE; i++)
		{
			if (Fitness[i] > nextBiggest)
			{
				nextBiggest = Fitness[i];
				index = i;
			}
		}
		Parent[p] = index;				// Remember where this parent was
		Fitness[index] = -1;			// Don't pick it again
	}
}

// Postfix incrementer--make the next generation of course:
template <class T>
int Generation<T>::operator++ (int)
{
	genCnt++;
	ChooseParents();

	int childCnt = 0;
	if ((genCnt % 2) == 0)
	{
		for (int i=0; i<PARENT_SIZE; i++)
		{
			for (int j=i+1; j<PARENT_SIZE; j++)
			{
				// Be fruitful and multiply:
				oddPopulation[childCnt++].MakeChild( 
					evenPopulation[Parent[i]],
					evenPopulation[Parent[j]]);
			}
		}
	}
	else
	{
		for (int i=0; i<PARENT_SIZE; i++)
		{
			for (int j=i+1; j<PARENT_SIZE; j++)
			{
				// Be fruitful and multiply:
				evenPopulation[childCnt++].MakeChild(
					oddPopulation[Parent[i]],
					oddPopulation[Parent[j]]);
			}
		}
	}
	// Oh no! Deadly radioactive fallout.  Aaagh, mutants!
	int pos;
	if ((genCnt % 2) == 0)
	{
		for (int i=0; i<mutationRate; i++)
		{
			pos = randomPlacement.Pick();
			int point = randomPoint.Pick();
			oddPopulation[randomMember.Pick()].
				m_Genes[point] = pos;
		}
	}
	else
	{
		for (int i=0; i<mutationRate; i++)
		{
			pos = randomPlacement.Pick();
			int point = randomPoint.Pick();
			evenPopulation[randomMember.Pick()].
				m_Genes[point] = pos;
		}
	}
	return bestOfBreed;
}

template <class T>
void Generation<T>::SetMutationRate(int rate)
{
	mutationRate = rate;
}

// Get the Best:
template <class T>
T Generation<T>::GetBest(void)
{
	return bestEver;
}

template <class T>
void Generation<T>::Init(const CArray<CPoint, CPoint&>& ptarray)
{
	pointCount = ptarray.GetSize();
	bestEverFitness = 0;
	randomPoint = Random<int>(0, pointCount-1);

	bestEver.Init(ptarray);
	for (int i=0; i<POP_SIZE; i++)
	{
		Fitness[i] = 0;
		oddPopulation[i].Init(ptarray);
		evenPopulation[i].Init(ptarray);
	}
}

#endif // !defined(AFX_GENERATION_H__8FA8C9D4_8B21_11D3_9887_0040056D374E__INCLUDED_)
